             .title        "Sunsoft-1 (INES 184) Plugin"

;05/05/07
;BootGod

;04/16/08
;BootGod
;Added check for CHR open bus

;10/04/09
;BootGod
;Fantasy Zone uses this chip for PRG banking instead of CHR. Uses presence of
;VRAM as indication of of what banking config is used. 

             ;vectors for standard system calls

send_byte:   .equ 0200h
baton:       .equ 0203h
chk_vram:    .equ 0206h
chk_wram:    .equ 0209h
wr_ppu:      .equ 020ch
read_byte:   .equ 020fh
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

tx:            .equ 00dfh
address:       .equ 00e0h
address_lo:    .equ 00e0h
address_hi:    .equ 00e1h
curbank:    .equ 00e2h
banksleft:    .equ 00e3h
bank0crc:  .equ 00e4h

reg:       .equ 00e8h
reg_lo:    .equ 00e8h
reg_hi:    .equ 00e9h

numpages_lo .equ 00eah
numpages_hi .equ 00ebh


temp_crc:    .equ 00ech

;plugin header defines
p_ver: = 3	;plugin version
p_read: = 1	;plugin is for downloading (dumping)
p_write: = 2	;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3	;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4	;plugin uses user variable #1
p_var2: = 8	;plugin uses user variable #2
p_var3: = 16	;plugin uses user variable #3
p_var4: = 32	;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "Sunsoft-1 Carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "BootGod"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 32 >> 4,128 >> 4
             .db 16 >> 3,32 >> 3
             .db 0,0

             ;plugin version and flags
             .db p_ver
             .db p_read

;plugin code start here

             
;determine H or V mirroring

             ldx #00h
             
             lda #020h
             jsr wr_ppu
             lda #055h
             sta 2007h
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #055h
             bne horz_mir
             lda 2007h
             cmp #0aah
             bne horz_mir
             lda #020h
             jsr wr_ppu
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #0aah
             bne horz_mir
             inx

horz_mir:    txa
             jsr send_byte   ;send byte

             jsr chk_vram
             bne prg32k

;if VRAM is present, PRG size is 128K

prg128k:     lda #0		;send PRG header
             jsr send_byte
             lda #2
             jsr send_byte
             lda #001h
             jsr send_byte

             lda #0
             sta address_lo
             sta curbank
             lda #8
             sta banksleft	;8 x 16K = 128K

set_pbank:   lda curbank
             sta 6000h
             inc curbank

             lda #080h
             sta address_hi	;set address to $8000

             ldy #0
             ldx #64		;setup for 16K bank dump


dump_pbank:  lda (address),y
             jsr send_byte
             iny
             bne dump_pbank
             inc address_hi
             dex
             bne dump_pbank
             dec banksleft
             bne set_pbank

             jmp done		;since was VRAM, nothing else to do

;if no VRAM, PRG size is 32K

prg32k:      lda #128		;send PRG header
             jsr send_byte
             lda #0
             jsr send_byte
             lda #001h
             jsr send_byte

             lda #0
             sta address_lo
             lda #080h
             sta address_hi	;set address to $8000

             ldy #0
             ldx #128		;setup for 32K bank dump

dump_prg:    lda (address),y
             jsr send_byte
             iny
             bne dump_prg
             inc address_hi
             dex
             bne dump_prg
             

;determine CHR ROM size

             lda #0
             jsr wr_cbank	;set to CHR bank 0
             jsr crc_cbank

             jsr ob_test	;check if first bank is open bus
             beq not_ob		;continue if it wasn't

             lda #64		;if it was open bus, setup for 16K dump
             sta numpages_lo	;starting from bank 4 and skip crc checks
             lda #0
             sta numpages_hi
             lda #4
             sta curbank
             sta banksleft
             jmp chr_hdr

not_ob:      ldx #3
ld_ccrc:     lda temp_crc,x	;save for comparision
             sta bank0crc,x
             dex
             bpl ld_ccrc
             
;start out at CHR bank 4
             lda #4
             sta curbank
             lda #0
             sta numpages_hi
             lda #64
             sta numpages_lo

next_ccrc:   lda curbank
             jsr wr_cbank	;set CHR bank num
             jsr crc_cbank
             
;compare crc's of bank 0 and curbank
             ldx #3
chk_ccrc:    lda bank0crc,x
             cmp temp_crc,x
             bne crc_cfail
             dex
             bpl chk_ccrc

             jmp got_chr

crc_cfail:   asl numpages_lo	;size in pages to send to pc
             asl curbank
             
got_chr:     lda curbank	;setup bank counters for dumping
             sta banksleft
             lda #0
             sta curbank

chr_hdr:     lda numpages_lo	;send CHR header
             jsr send_byte
             lda numpages_hi
             jsr send_byte
             lda #002h
             jsr send_byte
             
next_chr:    lda curbank
             jsr wr_cbank
             inc curbank
             lda #0
             jsr wr_ppu
             lda 2007h

             ldx #16
             ldy #0

dump_chr:    lda 2007h
             jsr send_byte
             iny
             bne dump_chr
             dex
             bne dump_chr
             dec banksleft
             bne next_chr

done:        lda #000h		;send termination header
             jsr send_byte
             lda #000h
             jsr send_byte
             lda #000h
             jsr send_byte

             rts

wr_cbank:    sta 6000h
             rts

;CRC routine for 4K CHR bank @ $0000, stores crc in temp_crc

crc_cbank:   lda #0
             jsr wr_ppu
             lda 2007h		;set PPU address to $0000

             jsr init_crc
             ldx #16
             ldy #0		;init crc for 4K of data (16 pages = 4K)

crc_cloop:   lda 2007h		;crc generation loop
             stx tx
             jsr do_crc
             ldx tx
             iny
             bne crc_cloop
             dex
             bne crc_cloop
             
             jsr finish_crc	;crc done

             ldx #3
crc_csave:   lda crc0,x		;save crc in temp_crc
             sta temp_crc,x
             dex
             bpl crc_csave

             rts

;Checks if last CRC was open bus

ob_test:     ldx #3		;check if open bus
ob_loop:     lda crc0,x
             cmp openbus,x
             bne crc_pass
             dex
             bne ob_loop

is_ob:       lda #1		;it is, send true
             rts

crc_pass:    lda #0		;it's not, send false
             rts

openbus:    .db 082h,020h,091h,0a2h


             .fill 0800h-*,0ffh   ;fill rest to get 1K of data

             .end
